from __future__ import absolute_import

import json
import re

try:
    from urllib.parse import urlencode
except ImportError:
    from urllib import urlencode

import tornado.gen
import tornado.web
import tornado.auth

from tornado.options import options
from celery.utils.imports import instantiate

from ..views import BaseHandler


class GoogleAuth2LoginHandler(BaseHandler, tornado.auth.GoogleOAuth2Mixin):
    _OAUTH_SETTINGS_KEY = 'oauth'

    @tornado.gen.coroutine
    def get(self):
        redirect_uri = self.settings[self._OAUTH_SETTINGS_KEY]['redirect_uri']
        if self.get_argument('code', False):
            user = yield self.get_authenticated_user(
                redirect_uri=redirect_uri,
                code=self.get_argument('code'),
            )
            yield self._on_auth(user)
        else:
            yield self.authorize_redirect(
                redirect_uri=redirect_uri,
                client_id=self.settings[self._OAUTH_SETTINGS_KEY]['key'],
                scope=['profile', 'email'],
                response_type='code',
                extra_params={'approval_prompt': 'auto'}
            )

    @tornado.gen.coroutine
    def _on_auth(self, user):
        if not user:
            raise tornado.web.HTTPError(403, 'Google auth failed')
        access_token = user['access_token']

        try:
            response = yield self.get_auth_http_client().fetch(
                'https://www.googleapis.com/userinfo/v2/me',
                headers={'Authorization': 'Bearer %s' % access_token})
        except Exception as e:
            raise tornado.web.HTTPError(403, 'Google auth failed: %s' % e)

        email = json.loads(response.body.decode('utf-8'))['email']
        if not re.match(self.application.options.auth, email):
            message = (
                "Access denied to '{email}'. Please use another account or "
                "ask your admin to add your email to flower --auth."
            ).format(email=email)
            raise tornado.web.HTTPError(403, message)

        self.set_secure_cookie("user", str(email))

        next_ = self.get_argument('next', self.application.options.url_prefix or '/')
        if self.application.options.url_prefix and next[0] != '/':
            next_ = '/' + next_

        self.redirect(next_)


class LoginHandler(BaseHandler):
    def __new__(cls, *args, **kwargs):
        return instantiate(options.auth_provider, *args, **kwargs)


class GithubLoginHandler(BaseHandler, tornado.auth.OAuth2Mixin):

    _OAUTH_AUTHORIZE_URL = "https://github.com/login/oauth/authorize"
    _OAUTH_ACCESS_TOKEN_URL = "https://github.com/login/oauth/access_token"
    _OAUTH_NO_CALLBACKS = False
    _OAUTH_SETTINGS_KEY = 'oauth'

    @tornado.gen.coroutine
    def get_authenticated_user(self, redirect_uri, code):
        body = urlencode({
            "redirect_uri": redirect_uri,
            "code": code,
            "client_id": self.settings[self._OAUTH_SETTINGS_KEY]['key'],
            "client_secret": self.settings[self._OAUTH_SETTINGS_KEY]['secret'],
            "grant_type": "authorization_code",
        })

        response = yield self.get_auth_http_client().fetch(
            self._OAUTH_ACCESS_TOKEN_URL,
            method="POST",
            headers={'Content-Type': 'application/x-www-form-urlencoded',
                     'Accept': 'application/json'}, body=body)

        if response.error:
            raise tornado.auth.AuthError(
                'OAuth authenticator error: %s' % str(response))

        raise tornado.gen.Return(json.loads(response.body.decode('utf-8')))

    @tornado.gen.coroutine
    def get(self):
        redirect_uri = self.settings[self._OAUTH_SETTINGS_KEY]['redirect_uri']
        if self.get_argument('code', False):
            user = yield self.get_authenticated_user(
                redirect_uri=redirect_uri,
                code=self.get_argument('code'),
            )
            yield self._on_auth(user)
        else:
            yield self.authorize_redirect(
                redirect_uri=redirect_uri,
                client_id=self.settings[self._OAUTH_SETTINGS_KEY]['key'],
                scope=['user:email'],
                response_type='code',
                extra_params={'approval_prompt': 'auto'}
            )

    @tornado.gen.coroutine
    def _on_auth(self, user):
        if not user:
            raise tornado.web.HTTPError(500, 'OAuth authentication failed')
        access_token = user['access_token']

        response = yield self.get_auth_http_client().fetch(
            'https://api.github.com/user/emails',
            headers={'Authorization': 'token ' + access_token,
                     'User-agent': 'Tornado auth'})


        emails = [email['email'].lower() for email in json.loads(response.body.decode('utf-8'))
                  if email['verified'] and re.match(self.application.options.auth, email['email'])]

        if not emails:
            message = (
                "Access denied. Please use another account or "
                "ask your admin to add your email to flower --auth."
            )
            raise tornado.web.HTTPError(403, message)

        self.set_secure_cookie("user", str(emails.pop()))

        next_ = self.get_argument('next', self.application.options.url_prefix or '/')
        if self.application.options.url_prefix and next_[0] != '/':
            next_ = '/' + next_
        self.redirect(next_)


class LogoutHandler(BaseHandler):
    def get(self):
        self.clear_cookie('user')
        self.render('404.html', message='Successfully logged out!')



from __future__ import absolute_import

import types

from celery import current_app
from django.conf import settings
from rpyc.utils.classic import DEFAULT_SERVER_PORT

from flower.events import Events


class Options(object):

    def __init__(self, namespace=None):
        self.namespace = namespace.upper()
        self.app = current_app

    def namespace_with(self, name):
        return self.namespace + "_" + name.upper()

    def define(self, name, default=None, **kwargs):
        value = getattr(settings, self.namespace_with(name), default)
        setattr(self, name, value)

    @property
    def state(self):
        return Events(self.app, self).get_remote_state()


options = Options('flower')

options.define("login_url", default=settings.LOGIN_URL, type=str,
               help="login url (settings.LOGIN_URL)")
options.define("inspect_timeout", default=1000, type=float,
               help="inspect timeout (in milliseconds)")
options.define("rpc_host", default='localhost', type=str,
               help="port used by rpc backend (int)")
options.define("rpc_port", default=DEFAULT_SERVER_PORT, type=int,
               help="port used by rpc backend (int)")
options.define("auth", default='', type=str,
               help="regexp of emails to grant access")
options.define("basic_auth", type=str, default=None, multiple=True,
               help="enable http basic authentication")
options.define("oauth2_key", type=str, default=None,
               help="OAuth2 key (requires --auth)")
options.define("oauth2_secret", type=str, default=None,
               help="OAuth2 secret (requires --auth)")
options.define("debug", default=settings.DEBUG,
               help="run in debug mode", type=bool)
options.define("oauth2_redirect_uri", type=str, default=None,
               help="OAuth2 redirect uri (requires --auth)")
options.define("max_workers", type=int, default=5000,
               help="maximum number of workers to keep in memory")
options.define("max_tasks", type=int, default=10000,
               help="maximum number of tasks to keep in memory")
options.define("db", type=str, default='flower',
               help="flower database file")
options.define("persistent", type=bool, default=False,
               help="enable persistent mode")
options.define("broker_api", type=str, default=None,
               help="inspect broker e.g. http://guest:guest@localhost:15672/api/")
options.define("ca_certs", type=str, default=None,
               help="SSL certificate authority (CA) file")
options.define("certfile", type=str, default=None,
               help="SSL certificate file")
options.define("keyfile", type=str, default=None,
               help="SSL key file")
options.define("xheaders", type=bool, default=False,
               help="enable support for the 'X-Real-Ip' and 'X-Scheme' headers.")
options.define("auto_refresh", default=True,
               help="refresh dashboards", type=bool)
options.define("cookie_secret", type=str, default=None,
               help="secure cookie secret")
options.define("enable_events", type=bool, default=True,
               help="periodically enable Celery events")
options.define("format_task", type=types.FunctionType, default=None,
               help="use custom task formatter")
options.define("natural_time", type=bool, default=False,
               help="show time in relative format")
options.define("tasks_columns", type=str,
               default="name,uuid,state,args,kwargs,result,received,started,runtime,worker",
               help="slugs of columns on /tasks/ page, delimited by comma")
options.define("auth_provider", default='flower.views.auth.GoogleAuth2LoginHandler',
               help="auth handler class")
options.define("inspect", default=False, help="inspect workers", type=bool)



from __future__ import absolute_import
from __future__ import with_statement

import collections
import logging
import time

import gevent.monkey
import rpyc
from celery.events import EventReceiver
from celery.events.state import State
from rpyc.utils import factory
from rpyc.utils.classic import DEFAULT_SERVER_PORT
from rpyc.utils.helpers import classpartial
from rpyc.utils.server import GeventServer

try:
    from collections import Counter
except ImportError:
    from .utils.backports.collections import Counter

logger = logging.getLogger(__name__)


class CeleryStateService(rpyc.Service):

    def __init__(self, state):
        super(CeleryStateService, self).__init__()
        self.state = state

    def exposed_get_state(self):
        return self.state

    def on_connect(self, conn):
        conn._config.update({
            'allow_public_attrs': True,
            'allow_pickle': True,
            'allow_all_attrs': True
        })


class EventsState(State):
    # EventsState object is created and accessed only from ioloop thread

    def __init__(self, *args, **kwargs):
        super(EventsState, self).__init__(*args, **kwargs)
        self.counter = collections.defaultdict(Counter)

    def event(self, event):
        worker_name = event['hostname']
        event_type = event['type']

        self.counter[worker_name][event_type] += 1

        # Send event to api subscribers (via websockets)
        # classname = api.events.getClassName(event_type)
        # cls = getattr(api.events, classname, None)
        # if cls:
        #     cls.send_message(event)

        # Save the event
        return super(EventsState, self).event(event)


class RpcClient(object):

    def __init__(self, service):
        self.service = service

    def connect(self, host, port=DEFAULT_SERVER_PORT, ipv6=False, keepalive=False):
        """
        Creates a socket connection to the given host and port.

        :param host: the host to connect to
        :param port: the TCP port
        :param ipv6: whether to create an IPv6 socket or IPv4

        :returns: an RPyC connection exposing ``Service``
        """
        return factory.connect(host, port, self.service, ipv6=ipv6, keepalive=keepalive)


class Events(object):

    rpc_conn = None

    def __init__(self, app, options):
        self.state = EventsState()
        self.options = options
        self.app = app
        self.server = None

        self.service = classpartial(CeleryStateService, self.state)
        self.client = RpcClient(self.service)

    def get_remote_state(self):
        if self.rpc_conn is None:
            self.rpc_conn = self.client.connect(self.options.rpc_host,
                                                port=self.options.rpc_port)
        return self.rpc_conn.root.get_state()

    def start_rpc(self):
        self.server = GeventServer(self.service,
                                   hostname=self.options.rpc_host,
                                   port=self.options.rpc_port,
                                   auto_register=False,
                                   logger=logger)
        self.server._listen()
        gevent.spawn(self.server.start)
        return self.server

    def run(self):
        self.start_rpc()
        try_interval = 1
        while True:
            try:
                try_interval *= 2

                with self.app.connection() as conn:
                    recv = EventReceiver(conn,
                                         handlers={"*": self.on_shutter},
                                         app=self.app)
                    try_interval = 1
                    recv.capture(limit=None, timeout=None, wakeup=True)
            except Exception as e:
                logger.error("Failed to capture events: '%s', "
                             "trying again in %s seconds.",
                             e, try_interval)
                logger.debug(e, exc_info=True)
                time.sleep(try_interval)

    def enable_events(self):
        # Periodically enable events for workers
        # launched after flower
        try:
            self.app.control.enable_events()
        except Exception as e:
            logger.debug("Failed to enable events: '%s'", e)

    def on_shutter(self, event):
        self.state.event(event)

        if not self.state.event_count:
            # No new events since last snapshot.
            return



from __future__ import absolute_import

import json
import logging
from datetime import datetime

from celery import states
from celery.backends.base import DisabledBackend
from celery.contrib.abortable import AbortableAsyncResult
from celery.result import AsyncResult
from flower.utils import login_required_admin
from django.utils.decorators import method_decorator

from flower.exceptions import HTTPError
from ..api.control import ControlHandler
from ..utils import tasks
from ..utils.broker import Broker
from ..views import BaseHandler

logger = logging.getLogger(__name__)


class BaseTaskHandler(BaseHandler):
    DATE_FORMAT = '%Y-%m-%d %H:%M:%S.%f'

    def get_task_args(self):
        try:
            body = self.request.body
            options = json.loads(body) if body else {}
        except ValueError as e:
            raise HTTPError(400, str(e))

        args = options.pop('args', [])
        kwargs = options.pop('kwargs', {})

        if not isinstance(args, (list, tuple)):
            raise HTTPError(400, 'args must be an array')

        return args, kwargs, options

    @staticmethod
    def backend_configured(result):
        return not isinstance(result.backend, DisabledBackend)

    def update_response_result(self, response, result):
        if result.state == states.FAILURE:
            response.update({'result': self.safe_result(result.result),
                             'traceback': result.traceback})
        else:
            response.update({'result': self.safe_result(result.result)})

    def normalize_options(self, options):
        if 'eta' in options:
            options['eta'] = datetime.strptime(options['eta'],
                                               self.DATE_FORMAT)
        if 'countdown' in options:
            options['countdown'] = float(options['countdown'])
        if 'expires' in options:
            expires = options['expires']
            try:
                expires = float(expires)
            except ValueError:
                expires = datetime.strptime(expires, self.DATE_FORMAT)
            options['expires'] = expires

    @staticmethod
    def safe_result(result):
        "returns json encodable result"
        try:
            json.dumps(result)
        except TypeError:
            return repr(result)
        else:
            return result


class TaskApply(BaseTaskHandler):

    @method_decorator(login_required_admin)
    def post(self, request, taskname):
        """
Execute a task by name and wait results

**Example request**:

.. sourcecode:: http

  POST /api/task/apply/tasks.add HTTP/1.1
  Accept: application/json
  Accept-Encoding: gzip, deflate, compress
  Content-Length: 16
  Content-Type: application/json; charset=utf-8
  Host: localhost:5555

  {
      "args": [1, 2]
  }

**Example response**:

.. sourcecode:: http

  HTTP/1.1 200 OK
  Content-Length: 71
  Content-Type: application/json; charset=UTF-8

  {
      "state": "SUCCESS",
      "task-id": "c60be250-fe52-48df-befb-ac66174076e6",
      "result": 3
  }

:query args: a list of arguments
:query kwargs: a dictionary of arguments
:reqheader Authorization: optional OAuth token to authenticate
:statuscode 200: no error
:statuscode 401: unauthorized request
:statuscode 404: unknown task
        """
        args, kwargs, options = self.get_task_args()
        logger.debug("Invoking a task '%s' with '%s' and '%s'",
                     taskname, args, kwargs)

        try:
            task = self.capp.tasks[taskname]
        except KeyError:
            raise HTTPError(404, "Unknown task '%s'" % taskname)

        try:
            self.normalize_options(options)
        except ValueError:
            raise HTTPError(400, 'Invalid option')

        result = task.apply_async(args=args, kwargs=kwargs, **options)
        response = {'task-id': result.task_id}

        return self.write(response)

    def wait_results(self, result, response):
        # Wait until task finished and do not raise anything
        result.get(propagate=False)
        # Write results and finish async function
        self.update_response_result(response, result)
        if self.backend_configured(result):
            response.update(state=result.state)
        return response


class TaskAsyncApply(BaseTaskHandler):
    DATE_FORMAT = '%Y-%m-%d %H:%M:%S.%f'

    @method_decorator(login_required_admin)
    def post(self, request, taskname):
        """
Execute a task

**Example request**:

.. sourcecode:: http

  POST /api/task/async-apply/tasks.add HTTP/1.1
  Accept: application/json
  Accept-Encoding: gzip, deflate, compress
  Content-Length: 16
  Content-Type: application/json; charset=utf-8
  Host: localhost:5555

  {
      "args": [1, 2]
  }

**Example response**:

.. sourcecode:: http

  HTTP/1.1 200 OK
  Content-Length: 71
  Content-Type: application/json; charset=UTF-8
  Date: Sun, 13 Apr 2014 15:55:00 GMT

  {
      "state": "PENDING",
      "task-id": "abc300c7-2922-4069-97b6-a635cc2ac47c"
  }

:query args: a list of arguments
:query kwargs: a dictionary of arguments
:query options: a dictionary of `apply_async` keyword arguments
:reqheader Authorization: optional OAuth token to authenticate
:statuscode 200: no error
:statuscode 401: unauthorized request
:statuscode 404: unknown task
        """
        args, kwargs, options = self.get_task_args()
        logger.debug("Invoking a task '%s' with '%s' and '%s'",
                     taskname, args, kwargs)

        try:
            task = self.capp.tasks[taskname]
        except KeyError:
            raise HTTPError(404, "Unknown task '%s'" % taskname)

        try:
            self.normalize_options(options)
        except ValueError:
            raise HTTPError(400, 'Invalid option')

        result = task.apply_async(args=args, kwargs=kwargs, **options)
        response = {'task-id': result.task_id}
        if self.backend_configured(result):
            response.update(state=result.state)
        return self.write(response)


class TaskSend(BaseTaskHandler):

    @method_decorator(login_required_admin)
    def post(self, request, taskname):
        """
Execute a task by name (doesn't require task sources)

**Example request**:

.. sourcecode:: http

  POST /api/task/send-task/tasks.add HTTP/1.1
  Accept: application/json
  Accept-Encoding: gzip, deflate, compress
  Content-Length: 16
  Content-Type: application/json; charset=utf-8
  Host: localhost:5555

  {
      "args": [1, 2]
  }

**Example response**:

.. sourcecode:: http

  HTTP/1.1 200 OK
  Content-Length: 71
  Content-Type: application/json; charset=UTF-8

  {
      "state": "SUCCESS",
      "task-id": "c60be250-fe52-48df-befb-ac66174076e6"
  }

:query args: a list of arguments
:query kwargs: a dictionary of arguments
:reqheader Authorization: optional OAuth token to authenticate
:statuscode 200: no error
:statuscode 401: unauthorized request
:statuscode 404: unknown task
        """
        args, kwargs, options = self.get_task_args()
        logger.debug("Invoking task '%s' with '%s' and '%s'",
                     taskname, args, kwargs)
        result = self.capp.send_task(
            taskname, args=args, kwargs=kwargs, **options)
        response = {'task-id': result.task_id}
        if self.backend_configured(result):
            response.update(state=result.state)
        return self.write(response)


class TaskResult(BaseTaskHandler):

    @method_decorator(login_required_admin)
    def get(self, request, taskid):
        """
Get a task result

**Example request**:

.. sourcecode:: http

  GET /api/task/result/c60be250-fe52-48df-befb-ac66174076e6 HTTP/1.1
  Host: localhost:5555

**Example response**:

.. sourcecode:: http

  HTTP/1.1 200 OK
  Content-Length: 84
  Content-Type: application/json; charset=UTF-8

  {
      "result": 3,
      "state": "SUCCESS",
      "task-id": "c60be250-fe52-48df-befb-ac66174076e6"
  }

:query timeout: how long to wait, in seconds, before the operation times out
:reqheader Authorization: optional OAuth token to authenticate
:statuscode 200: no error
:statuscode 401: unauthorized request
:statuscode 503: result backend is not configured
        """
        timeout = self.get_argument('timeout', None)
        timeout = float(timeout) if timeout is not None else None

        result = AsyncResult(taskid)
        if not self.backend_configured(result):
            raise HTTPError(503)

        response = {'task-id': taskid, 'state': result.state}

        if timeout:
            result.get(timeout=timeout, propagate=False)
            self.update_response_result(response, result)
        elif result.ready():
            self.update_response_result(response, result)

        return self.write(response)


class TaskAbort(BaseTaskHandler):

    @method_decorator(login_required_admin)
    def post(self, request, taskid):
        """
Abort a running task

**Example request**:

.. sourcecode:: http

  POST /api/task/abort/c60be250-fe52-48df-befb-ac66174076e6 HTTP/1.1
  Host: localhost:5555

**Example response**:

.. sourcecode:: http

  HTTP/1.1 200 OK
  Content-Length: 61
  Content-Type: application/json; charset=UTF-8

  {
      "message": "Aborted '1480b55c-b8b2-462c-985e-24af3e9158f9'"
  }

:reqheader Authorization: optional OAuth token to authenticate
:statuscode 200: no error
:statuscode 401: unauthorized request
:statuscode 503: result backend is not configured
        """
        logger.info("Aborting task '%s'", taskid)

        result = AbortableAsyncResult(taskid)
        if not self.backend_configured(result):
            raise HTTPError(503)

        result.abort()

        return self.write(dict(message="Aborted '%s'" % taskid))


class GetQueueLengths(BaseTaskHandler):

    @method_decorator(login_required_admin)
    def get(self, request):
        """
Return length of all active queues

**Example request**:

.. sourcecode:: http

  GET /api/queues/length
  Host: localhost:5555

**Example response**:

.. sourcecode:: http

  HTTP/1.1 200 OK
  Content-Length: 94
  Content-Type: application/json; charset=UTF-8

  {
      "active_queues": [
          {"name": "celery", "messages": 0},
          {"name": "video-queue", "messages": 5}
      ]
  }

:reqheader Authorization: optional OAuth token to authenticate
:statuscode 200: no error
:statuscode 401: unauthorized request
:statuscode 503: result backend is not configured
        """
        app = self.capp
        broker_options = app.conf.BROKER_TRANSPORT_OPTIONS
        http_api = None

        from celery.backends.amqp import AMQPBackend
        if isinstance(app.backend, AMQPBackend) == 'amqp' and self.settings.broker_api:
            http_api = app.options.broker_api

        broker = Broker(app.connection().as_uri(include_password=True),
                        http_api=http_api, broker_options=broker_options)

        queue_names = ControlHandler.get_active_queue_names()

        if not queue_names:
            queue_names = set([app.conf.CELERY_DEFAULT_QUEUE]) |\
                          set([q.name for q in app.conf.CELERY_QUEUES or [] if q.name])

        queues = broker.queues(sorted(queue_names))
        return self.write({'active_queues': queues})


class ListTasks(BaseTaskHandler):

    @method_decorator(login_required_admin)
    def get(self, request):
        """
List tasks

**Example request**:

.. sourcecode:: http

  GET /api/tasks HTTP/1.1
  Host: localhost:5555
  User-Agent: HTTPie/0.8.0

**Example response**:

.. sourcecode:: http

  HTTP/1.1 200 OK
  Content-Length: 1109
  Content-Type: application/json; charset=UTF-8
  Etag: "b2478118015c8b825f7b88ce6b660e5449746c37"
  Server: TornadoServer/3.1.1

  {
      "e42ceb2d-8730-47b5-8b4d-8e0d2a1ef7c9": {
          "args": "[3, 4]",
          "client": null,
          "clock": 1079,
          "eta": null,
          "exception": null,
          "exchange": null,
          "expires": null,
          "failed": null,
          "kwargs": "{}",
          "name": "tasks.add",
          "received": 1398505411.107885,
          "result": "'7'",
          "retried": null,
          "retries": 0,
          "revoked": null,
          "routing_key": null,
          "runtime": 0.01610181899741292,
          "sent": null,
          "started": 1398505411.108985,
          "state": "SUCCESS",
          "succeeded": 1398505411.124802,
          "timestamp": 1398505411.124802,
          "traceback": null,
          "uuid": "e42ceb2d-8730-47b5-8b4d-8e0d2a1ef7c9"
      },
      "f67ea225-ae9e-42a8-90b0-5de0b24507e0": {
          "args": "[1, 2]",
          "client": null,
          "clock": 1042,
          "eta": null,
          "exception": null,
          "exchange": null,
          "expires": null,
          "failed": null,
          "kwargs": "{}",
          "name": "tasks.add",
          "received": 1398505395.327208,
          "result": "'3'",
          "retried": null,
          "retries": 0,
          "revoked": null,
          "routing_key": null,
          "runtime": 0.012884548006695695,
          "sent": null,
          "started": 1398505395.3289,
          "state": "SUCCESS",
          "succeeded": 1398505395.341089,
          "timestamp": 1398505395.341089,
          "traceback": null,
          "uuid": "f67ea225-ae9e-42a8-90b0-5de0b24507e0"
      }
  }

:query limit: maximum number of tasks
:query workername: filter task by workername
:query taskname: filter tasks by taskname
:query state: filter tasks by state
:query received_start: filter tasks by received date (must be greater than) format %Y-%m-%d %H:%M
:query received_end: filter tasks by received date (must be less than) format %Y-%m-%d %H:%M
:reqheader Authorization: optional OAuth token to authenticate
:statuscode 200: no error
:statuscode 401: unauthorized request
        """
        app_state = self.settings.state

        limit = self.get_argument('limit', None)
        worker = self.get_argument('workername', None)
        type = self.get_argument('taskname', None)
        state = self.get_argument('state', None)
        received_start = self.get_argument('received_start', None)
        received_end = self.get_argument('received_end', None)

        limit = limit and int(limit)
        worker = worker if worker != 'All' else None
        type = type if type != 'All' else None
        state = state if state != 'All' else None

        result = []
        for task_id, task in tasks.iter_tasks(
                app_state, limit=limit, type=type,
                worker=worker, state=state,
                received_start=received_start,
                received_end=received_end):
            task = tasks.as_dict(task)
            task.pop('worker', None)
            result.append((task_id, task))

        return self.write(dict(result))


class ListTaskTypes(BaseTaskHandler):

    @method_decorator(login_required_admin)
    def get(self, request, *args, **kwargs):
        """
List (seen) task types

**Example request**:

.. sourcecode:: http

  GET /api/task/types HTTP/1.1
  Host: localhost:5555

**Example response**:

.. sourcecode:: http

  HTTP/1.1 200 OK
  Content-Length: 44
  Content-Type: application/json; charset=UTF-8

  {
      "task-types": [
          "tasks.add",
          "tasks.sleep"
      ]
  }

:reqheader Authorization: optional OAuth token to authenticate
:statuscode 200: no error
:statuscode 401: unauthorized request
        """
        seen_task_types = self.settings.state.task_types()
        response = {'task-types': seen_task_types}
        return self.write(response)


class TaskInfo(BaseTaskHandler):

    @method_decorator(login_required_admin)
    def get(self, request, taskid):
        """
Get a task info

**Example request**:

.. sourcecode:: http

  GET /api/task/info/91396550-c228-4111-9da4-9d88cfd5ddc6 HTTP/1.1
  Accept: */*
  Accept-Encoding: gzip, deflate, compress
  Host: localhost:5555


**Example response**:

.. sourcecode:: http

  HTTP/1.1 200 OK
  Content-Length: 575
  Content-Type: application/json; charset=UTF-8

  {
      "args": "[2, 2]",
      "client": null,
      "clock": 25,
      "eta": null,
      "exception": null,
      "exchange": null,
      "expires": null,
      "failed": null,
      "kwargs": "{}",
      "name": "tasks.add",
      "received": 1400806241.970742,
      "result": "'4'",
      "retried": null,
      "retries": null,
      "revoked": null,
      "routing_key": null,
      "runtime": 2.0037889280356467,
      "sent": null,
      "started": 1400806241.972624,
      "state": "SUCCESS",
      "succeeded": 1400806243.975336,
      "task-id": "91396550-c228-4111-9da4-9d88cfd5ddc6",
      "timestamp": 1400806243.975336,
      "traceback": null,
      "worker": "celery@worker1"
  }

:reqheader Authorization: optional OAuth token to authenticate
:statuscode 200: no error
:statuscode 401: unauthorized request
:statuscode 404: unknown task
        """

        task = tasks.get_task_by_id(self.settings.state, taskid)
        if not task:
            raise HTTPError(404, "Unknown task '%s'" % taskid)

        response = task.as_dict()
        if task.worker is not None:
            response['worker'] = task.worker.hostname

        return self.write(response)



from __future__ import absolute_import

import copy
import logging
from functools import total_ordering

from flower.utils import login_required_admin
from django.utils.decorators import method_decorator

from flower.exceptions import HTTPError

try:
    from itertools import imap
except ImportError:
    imap = map

from ..views import BaseHandler
from ..utils.tasks import iter_tasks, get_task_by_id, as_dict

logger = logging.getLogger(__name__)


class TaskView(BaseHandler):

    @method_decorator(login_required_admin)
    def get(self, request, task_id):
        try:
            task = get_task_by_id(self.settings.state, task_id)
        except Exception:
            raise HTTPError(404, "Unknown task '%s'" % task_id)

        return self.render("flower/task.html", context={'task': task})


@total_ordering
class Comparable(object):
    """
    Compare two objects, one or more of which may be None.  If one of the
    values is None, the other will be deemed greater.
    """

    def __init__(self, value):
        self.value = value

    def __eq__(self, other):
        return self.value == other.value

    def __lt__(self, other):
        try:
            return self.value < other.value
        except TypeError:
            return self.value is None


class TasksDataTable(BaseHandler):

    @method_decorator(login_required_admin)
    def get(self, request):
        state = self.settings.state

        draw = self.get_argument('draw', type=int)
        start = self.get_argument('start', type=int)
        length = self.get_argument('length', type=int)
        search = self.get_argument('search[value]', type=str)

        column = self.get_argument('order[0][column]', type=int)
        sort_by = self.get_argument('columns[%s][data]' % column, type=str)
        sort_order = self.get_argument('order[0][dir]', type=str) == 'desc'

        def key(item):
            return Comparable(getattr(item[1], sort_by))

        sorted_tasks = sorted(
            iter_tasks(state, search=search),
            key=key,
            reverse=sort_order
        )

        filtered_tasks = []
        for task in sorted_tasks[start:start + length]:
            task_dict = as_dict(self.format_task(task)[1])
            if task_dict.get('worker'):
                task_dict['worker'] = task_dict['worker'].hostname

            filtered_tasks.append(task_dict)

        return self.write(dict(draw=draw, data=filtered_tasks,
                          recordsTotal=len(sorted_tasks),
                          recordsFiltered=len(sorted_tasks)))

    @method_decorator(login_required_admin)
    def post(self, request):
        return self.get(request)

    def format_task(self, args):
        uuid, task = args
        custom_format_task = self.settings.format_task
        if custom_format_task:
            try:
                task = custom_format_task(copy.copy(task))
            except:
                logger.exception("Failed to format '%s' task", uuid)
        return uuid, task


class TasksView(BaseHandler):

    @method_decorator(login_required_admin)
    def get(self, request, *args, **kwargs):
        settings = self.settings
        app = settings.app
        time = 'natural-time' if settings.natural_time else 'time'
        if app.conf.CELERY_TIMEZONE:
            time += '-' + str(app.conf.CELERY_TIMEZONE)
        # state_config = self.get_argument("state", default=None)
        context = dict(
            tasks=[],
            columns=settings.tasks_columns,
            time=time,
        )
        return self.render("flower/tasks.html", context=context)



from __future__ import absolute_import

import sys
import json
import socket
import logging
import numbers

import requests

try:
    from urllib.parse import urlparse, urljoin, quote, unquote
except ImportError:
    from urlparse import urlparse, urljoin
    from urllib import quote, unquote


try:
    import redis
except ImportError:
    redis = None


logger = logging.getLogger(__name__)


class BrokerBase(object):
    def __init__(self, broker_url, *args, **kwargs):
        purl = urlparse(broker_url)
        self.host = purl.hostname
        self.port = purl.port
        self.vhost = purl.path[1:]

        username = purl.username
        password = purl.password

        self.username = unquote(username) if username else username
        self.password = unquote(password) if password else password

    def queues(self, names):
        raise NotImplementedError


class RabbitMQ(BrokerBase):
    def __init__(self, broker_url, http_api, io_loop=None, **kwargs):
        super(RabbitMQ, self).__init__(broker_url)

        self.host = self.host or 'localhost'
        self.port = self.port or 15672
        self.vhost = quote(self.vhost, '') or '/'
        self.username = self.username or 'guest'
        self.password = self.password or 'guest'

        if not http_api:
            http_api = "http://{username}:{password}@{host}:{port}/api/{vhost}".format(
                username=self.username, password=self.password,
                host=self.host, port=self.port, vhost=self.vhost
            )

        try:
            self.validate_http_api(http_api)
        except Exception as e:
            logger.error("Invalid broker api url:%s", http_api)

        self.http_api = http_api

    def queues(self, names):
        url = urljoin(self.http_api, 'queues/' + self.vhost)
        api_url = urlparse(self.http_api)
        username = unquote(api_url.username or '') or self.username
        password = unquote(api_url.password or '') or self.password

        try:
            response = requests.get(
                url,
                auth_username=username,
                auth_password=password,
                validate_cert=False
            )
        except (socket.error, requests.HTTPError) as e:
            logger.error("RabbitMQ management API call failed: %s", e)
            raise

        if response.status_code == 200:
            info = response.json()
            return [x for x in info if x['name'] in names]
        else:
            raise Exception('Status({0.status_code})'.format(response))

    @classmethod
    def validate_http_api(cls, http_api):
        url = urlparse(http_api)
        if url.scheme not in ('http', 'https'):
            raise ValueError("Invalid http api schema: %s" % url.scheme)


class RedisBase(BrokerBase):
    SEP = '\x06\x16'
    DEFAULT_PRIORITY_STEPS = [0, 3, 6, 9]

    def __init__(self, broker_url, *args, **kwargs):
        super(RedisBase, self).__init__(broker_url)

        if not redis:
            raise ImportError('redis library is required')

        broker_options = kwargs.get('broker_options')

        if broker_options and 'priority_steps' in broker_options:
            self.priority_steps = broker_options['priority_steps']
        else:
            self.priority_steps = self.DEFAULT_PRIORITY_STEPS

    def _q_for_pri(self, queue, pri):
        if pri not in self.priority_steps:
            raise ValueError('Priority not in priority steps')
        return '{0}{1}{2}'.format(*((queue, self.SEP, pri) if pri else (queue, '', '')))

    def queues(self, names):
        queue_stats = []
        for name in names:
            priority_names = [self._q_for_pri(name, pri) for pri in self.priority_steps]
            queue_stats.append({
                'name': name,
                'messages': sum([self.redis.llen(x) for x in priority_names])
            })
        return queue_stats


class Redis(RedisBase):

    def __init__(self, broker_url, *args, **kwargs):
        super(Redis, self).__init__(broker_url)
        self.host = self.host or 'localhost'
        self.port = self.port or 6379
        self.vhost = self._prepare_virtual_host(self.vhost)

        self.redis = redis.Redis(host=self.host, port=self.port,
                                 db=self.vhost, password=self.password)

    def _prepare_virtual_host(self, vhost):
        if not isinstance(vhost, numbers.Integral):
            if not vhost or vhost == '/':
                vhost = 0
            elif vhost.startswith('/'):
                vhost = vhost[1:]
            try:
                vhost = int(vhost)
            except ValueError:
                raise ValueError(
                    'Database is int between 0 and limit - 1, not {0}'.format(
                        vhost,
                    ))
        return vhost


class RedisSocket(RedisBase):

    def __init__(self, broker_url, *args, **kwargs):
        super(RedisSocket, self).__init__(broker_url)
        self.redis = redis.Redis(unix_socket_path='/' + self.vhost,
                                 password=self.password)


class Broker(object):
    def __new__(cls, broker_url, *args, **kwargs):
        scheme = urlparse(broker_url).scheme
        if scheme == 'amqp':
            return RabbitMQ(broker_url, *args, **kwargs)
        elif scheme == 'redis':
            return Redis(broker_url, *args, **kwargs)
        elif scheme == 'redis+socket':
            return RedisSocket(broker_url, *args, **kwargs)
        else:
            raise NotImplementedError




import time

from tests.unit import AsyncHTTPTestCase


class MonitorTest(AsyncHTTPTestCase):
    def test_monitor_page(self):
        r = self.get('/monitor')
        self.assertEqual(200, r.code)
        self.assertTrue('Succeeded tasks' in str(r.body))
        self.assertTrue('Failed tasks' in str(r.body))

    def test_monitor_succeeded_tasks(self):
        r = self.get('/monitor/succeeded-tasks?lastquery=%s' % time.time())
        self.assertEqual(200, r.code)

    def test_monitor_completion_time(self):
        r = self.get('/monitor/completion-time?lastquery=%s' % time.time())
        self.assertEqual(200, r.code)

    def test_monitor_failed_tasks(self):
        r = self.get('/monitor/failed-tasks?lastquery=%s' % time.time())
        self.assertEqual(200, r.code)



from __future__ import absolute_import

import copy
import inspect
import re
import traceback
from base64 import b64decode
from distutils.util import strtobool

from django.conf import settings
from django.http import Http404, JsonResponse, HttpResponse
from django.shortcuts import render
from django.urls import reverse
from django.utils.functional import cached_property
from django.utils.six import string_types
from django.views.generic import View
from rpyc import BaseNetref

from flower.exceptions import HTTPError
from flower.options import options
from flower.utils import template
from ..utils import bugreport, prepend_url


class BaseHandler(View):
    template_engine = 'jinja2'

    def __init__(self, *args, **kwargs):
        super(BaseHandler, self).__init__(*args, **kwargs)
        self.url_prefix = reverse("flower:main").rstrip(' / ')
        self.settings = options

    def render(self, template_name, context=None, **kwargs):
        if context is None:
            context = {}
        functions = inspect.getmembers(template, inspect.isfunction)
        assert not set(map(lambda x: x[0], functions)) & set(context.keys())
        context.update(functions)
        context.update({
            'url_prefix': self.url_prefix,
            'isinstance': isinstance
        })
        return render(self.request, template_name,
                      context=context,
                      using=self.template_engine,
                      **kwargs)

    @staticmethod
    def json_default(o):
        """netref as local data"""
        if isinstance(o, BaseNetref):
            o = copy.deepcopy(o)
        return o

    def write(self, data):
        return JsonResponse(data, safe=False, json_dumps_params={
            'default': self.json_default
        })

    def write_error(self, status_code, **kwargs):
        error_trace = ""
        if 'exc_info' in kwargs:
            for line in traceback.format_exception(*kwargs['exc_info']):
                error_trace += line

        if status_code in (404, 403):
            message = kwargs.get('message')
            response = self.render('flower/404.html', context=dict(
                message=message,
                error_trace=error_trace
            ), status=status_code)
        elif status_code == 500:
            message = kwargs.get('message')
            response = self.render('flower/error.html', context=dict(
                debug=self.settings.debug,
                status_code=status_code,
                error_trace=error_trace,
                message=message
            ), status=status_code)

        elif status_code == 401:
            response = HttpResponse('Access denied', status=status_code)
            response['WWW-Authenticate'] = 'Basic realm="flower"'
        else:
            response = HttpResponse(status=status_code)
            if 'exc_info' in kwargs and kwargs['exc_info'][0] == Http404:
                message = kwargs['exc_info'][1].log_message
                response['Content-Type'] = 'text/plain'
                response.write(message)
        return response

    def get_current_user(self):
        # Basic Auth
        basic_auth = self.settings.basic_auth
        if basic_auth:
            auth_header = self.request.headers.get("Authorization", "")
            try:
                basic, credentials = auth_header.split()
                credentials = b64decode(credentials.encode()).decode()
                if basic != 'Basic' or credentials not in basic_auth:
                    raise HTTPError(401)
            except ValueError:
                raise HTTPError(401)

        # Google OpenID
        if not self.settings.auth:
            return True

        user = self.request.user
        username = getattr(user, settings.USERNAME_FIELD)
        if re.search(self.settings.auth, username):
            return user
        return None

    def get_argument(self, name, default=None, strip=True, type=None):
        arg = getattr(self.request, self.request.method).get(name, default)
        if isinstance(arg, string_types) and strip:
            arg = arg.strip()
        if type is not None:
            try:
                if type is bool:
                    arg = type(strtobool(str(arg)))
                else:
                    arg = type(arg)
            except (ValueError, TypeError):
                if arg is None and default is None:
                    return arg
                raise Http404("Invalid argument '%s' of type '%s'" % (
                    arg, type.__name__))
        return arg

    @cached_property
    def capp(self):
        """return Celery application object"""
        return self.settings.app

    def reverse_url(self, *args):
        prefix = self.settings.url_prefix
        url = super(BaseHandler, self).reverse_url(*args)
        return prepend_url(url, prefix) if prefix else url



# coding=utf-8

import gevent.monkey

gevent.monkey.patch_all()

from django.core.management import BaseCommand

from flower.events import Events, logger
from flower.options import options as settings


class Command(BaseCommand):
    # db can not be used
    leave_locale_alone = True
    requires_system_checks = False

    def handle(self, *args, **options):
        events = Events(settings.app, settings)
        try:
            settings.app.control.enable_events()
        except Exception as e:
            logger.debug("Failed to enable events: '%s'", e)
        events.run()



from __future__ import absolute_import

import logging
import socket
import sys
from collections import OrderedDict
from functools import partial

from flower.utils import login_required_admin
from django.http import JsonResponse
from django.utils.decorators import method_decorator
from django.views.generic import View

from ..api.workers import ListWorkers
from ..views import BaseHandler

logger = logging.getLogger(__name__)


class DashboardView(BaseHandler):

    @method_decorator(login_required_admin)
    def get(self, request, *args, **kwargs):
        refresh = self.get_argument('refresh', default=False, type=bool)
        json = self.get_argument('json', default=False, type=bool)
        app = self.settings.app

        # When rpc fails to connect (check the flower_events command is running).
        try:
            events = self.settings.state
        except socket.error:
            return self.write_error(500, message='RPC connection failed',
                                    exc_info=sys.exc_info())

        if refresh:
            try:
                return JsonResponse(list(ListWorkers.update_workers(settings=self.settings)))
            except Exception as e:
                logger.exception('Failed to update workers: %s', e)

        workers = {}
        for name, values in events.counter.items():
            if name not in events.workers:
                continue
            worker = events.workers[name]
            info = dict(values)
            info.update(self._as_dict(worker))
            info.update(status=worker.alive)
            workers[name] = info

        if json:
            return self.write(dict(data=list(workers.values())))
        else:
            broker = app.connection().as_uri()

            def lazy_alive_workers():
                return sum(map(lambda x: x.get('active') or 0, workers.values()))

            def lazy_task_received():
                return sum(map(lambda x: x.get('task-received') or 0, workers.values()))

            def lazy_task_failed():
                return sum(map(lambda x: x.get('task-failed') or 0, workers.values()))

            def lazy_task_succeeded():
                return sum(map(lambda x: x.get('task-succeeded') or 0, workers.values()))

            def lazy_task_retried():
                return sum(map(lambda x: x.get('task-retried') or 0, workers.values()))

            context = dict(
                alive_workers=lazy_alive_workers,
                task_received=lazy_task_received,
                task_failed=lazy_task_failed,
                task_succeeded=lazy_task_succeeded,
                task_retried=lazy_task_retried,
                broker=broker,
                workers=workers
            )
            return self.render("flower/dashboard.html", context)

    @classmethod
    def _as_dict(cls, worker):
        if hasattr(worker, '_fields'):
            return dict((k, worker.__getattribute__(k)) for k in worker._fields)
        else:
            return cls._info(worker)

    @classmethod
    def _info(cls, worker):
        _fields = ('hostname', 'pid', 'freq', 'heartbeats', 'clock',
                   'active', 'processed', 'loadavg', 'sw_ident',
                   'sw_ver', 'sw_sys')

        def _keys():
            for key in _fields:
                value = getattr(worker, key, None)
                if value is not None:
                    yield key, value

        return dict(_keys())


class DashboardUpdateHandler(View):
    listeners = []
    periodic_callback = None
    workers = None
    page_update_interval = 2000

    def open(self):
        app = self.app_options
        if not app.auto_refresh:
            self.write_message({})
            return

        if not self.listeners:
            if self.periodic_callback is None:
                cls = DashboardUpdateHandler
                cls.periodic_callback = PeriodicCallback(
                    partial(cls.on_update_time, app),
                    self.page_update_interval)
            if not self.periodic_callback._running:
                logger.debug('Starting a timer for dashboard updates')
                self.periodic_callback.start()
        self.listeners.append(self)

    def on_message(self, message):
        pass

    def on_close(self):
        if self in self.listeners:
            self.listeners.remove(self)
        if not self.listeners and self.periodic_callback:
            logger.debug('Stopping dashboard updates timer')
            self.periodic_callback.stop()

    @classmethod
    def on_update_time(cls, app):
        update = cls.dashboard_update(app)
        if update:
            for l in cls.listeners:
                l.write_message(update)

    @classmethod
    def dashboard_update(cls, app):
        state = app.events.state
        workers = OrderedDict()

        for name, worker in sorted(state.workers.items()):
            counter = state.counter[name]
            started = counter.get('task-started', 0)
            processed = counter.get('task-received', 0)
            failed = counter.get('task-failed', 0)
            succeeded = counter.get('task-succeeded', 0)
            retried = counter.get('task-retried', 0)
            active = started - succeeded - failed - retried
            if active < 0:
                active = 'N/A'

            workers[name] = dict(
                name=name,
                status=worker.alive,
                active=active,
                processed=processed,
                failed=failed,
                succeeded=succeeded,
                retried=retried,
                loadavg=getattr(worker, 'loadavg', None))
        return workers

    def check_origin(self, origin):
        return True



from __future__ import absolute_import

import logging

from flower.utils import login_required_admin
from django.utils.decorators import method_decorator

from ..api.workers import ListWorkers
from ..views import BaseHandler

logger = logging.getLogger(__name__)


class WorkerView(BaseHandler):

    @method_decorator(login_required_admin)
    def get(self, request, name):
        try:
            ListWorkers.update_workers(settings=self.settings,
                                       workername=name)
        except Exception as e:
            logger.error(e)

        worker = ListWorkers.worker_cache.get(name)

        if worker is None:
            return self.write_error(404, message="Unknown worker '%s'" % name)

        if 'stats' not in worker:
            return self.write_error(404, message="Unable to get stats for '%s' worker" % name)

        context = {
            'worker': dict(worker, name=name)
        }
        return self.render("flower/worker.html", context=context)



import unittest

from mock import MagicMock

from flower.utils import broker
from flower.utils.broker import RabbitMQ, Redis, RedisSocket, Broker


broker.requests = MagicMock()
broker.redis = MagicMock()


class TestRabbitMQ(unittest.TestCase):
    def test_init(self):
        b = Broker('amqp://', '')
        self.assertTrue(isinstance(b, RabbitMQ))
        self.assertFalse(isinstance(b, Redis))

    def test_url(self):
        b = RabbitMQ('amqp://user:pass@host:10000/vhost', '')
        self.assertEqual('host', b.host)
        self.assertEqual(10000, b.port)
        self.assertEqual('vhost', b.vhost)
        self.assertEqual('user', b.username)
        self.assertEqual('pass', b.password)

    def test_url_defaults_rabbitmq(self):
        for url in ['amqp://', 'amqp://localhost']:
            b = RabbitMQ(url, '')
            self.assertEqual('localhost', b.host)
            self.assertEqual(15672, b.port)
            self.assertEqual('/', b.vhost)
            self.assertEqual('guest', b.username)
            self.assertEqual('guest', b.password)

    def test_url_defaults_redis(self):
        for url in ['redis://', 'redis://localhost', 'redis://localhost/0']:
            b = Redis(url, '')
            self.assertEqual('localhost', b.host)
            self.assertEqual(6379, b.port)
            self.assertEqual(0, b.vhost)
            self.assertEqual(None, b.username)
            self.assertEqual(None, b.password)


class TestRedis(unittest.TestCase):
    def test_init(self):
        b = Broker('redis://localhost:6379/0')
        self.assertFalse(isinstance(b, RabbitMQ))
        self.assertTrue(isinstance(b, Redis))

    def test_url(self):
        b = Broker('redis://foo:7777/9')
        self.assertEqual('foo', b.host)
        self.assertEqual(7777, b.port)
        self.assertEqual(9, b.vhost)

    def test_url_defaults(self):
        b = Broker('redis://')
        self.assertEqual('localhost', b.host)
        self.assertEqual(6379, b.port)
        self.assertEqual(0, b.vhost)
        self.assertIsNone(b.username)
        self.assertIsNone(b.password)

    def test_url_with_password(self):
        b = Broker('redis://:pass@host:4444/5')
        self.assertEqual('host', b.host)
        self.assertEqual(4444, b.port)
        self.assertEqual(5, b.vhost)
        self.assertEqual('pass', b.password)


class TestRedisSocket(unittest.TestCase):
    def test_init(self):
        b = Broker('redis+socket:///path/to/socket')
        self.assertFalse(isinstance(b, RabbitMQ))
        self.assertTrue(isinstance(b, RedisSocket))

    def test_url(self):
        b = Broker('redis+socket:///path/to/socket')
        self.assertEqual(None, b.host)
        self.assertEqual(None, b.port)
        self.assertEqual('path/to/socket', b.vhost)


if __name__ == '__main__':
    unittest.main()



from __future__ import absolute_import

from django.conf.urls import url, include
from django.views.decorators.cache import never_cache

from .api import control
from .api import tasks
from .api import workers
from .views import monitor
from .views.broker import BrokerView
from .views.dashboard import DashboardView
from .views.error import NotFoundErrorHandler
from .views.tasks import TaskView, TasksView, TasksDataTable
from .views.workers import WorkerView


ns_urlpatterns = ([
    # App
    url(r"^$", never_cache(DashboardView.as_view()), name='main'),
    url(r"^dashboard/$", never_cache(DashboardView.as_view()), name='dashboard'),
    url(r"^worker/(.+)", WorkerView.as_view(), name='worker'),
    url(r"^task/(.+)", TaskView.as_view(), name='task'),
    url(r"^tasks/$", never_cache(TasksView.as_view()), name='tasks'),
    url(r"^tasks/datatable/$", TasksDataTable.as_view()),
    url(r"^broker/$", BrokerView.as_view(), name='broker'),

    # # Worker API
    url(r"^api/workers/$", workers.ListWorkers.as_view()),
    url(r"^api/worker/shutdown/(.+)", control.WorkerShutDown.as_view()),
    url(r"^api/worker/pool/restart/(.+)", control.WorkerPoolRestart.as_view()),
    url(r"^api/worker/pool/grow/(.+)", control.WorkerPoolGrow.as_view()),
    url(r"^api/worker/pool/shrink/(.+)", control.WorkerPoolShrink.as_view()),
    url(r"^api/worker/pool/autoscale/(.+)", control.WorkerPoolAutoscale.as_view()),
    url(r"^api/worker/queue/add-consumer/(.+)", control.WorkerQueueAddConsumer.as_view()),
    url(r"^api/worker/queue/cancel-consumer/(.+)", control.WorkerQueueCancelConsumer.as_view()),

    # Task API
    url(r"^api/tasks/$", tasks.ListTasks.as_view()),
    url(r"^api/task/types/$", tasks.ListTaskTypes.as_view()),
    url(r"^api/queues/length/$", tasks.GetQueueLengths.as_view()),
    url(r"^api/task/info/(.*)", tasks.TaskInfo.as_view()),
    url(r"^api/task/apply/(.+)", tasks.TaskApply.as_view()),
    url(r"^api/task/async-apply/(.+)", tasks.TaskAsyncApply.as_view()),
    url(r"^api/task/send-task/(.+)", tasks.TaskSend.as_view()),
    url(r"^api/task/result/(.+)", tasks.TaskResult.as_view()),
    url(r"^api/task/abort/(.+)", tasks.TaskAbort.as_view()),
    url(r"^api/task/timeout/(.+)", control.TaskTimout.as_view()),
    url(r"^api/task/rate-limit/(.+)", control.TaskRateLimit.as_view()),
    url(r"^api/task/revoke/(.+)", control.TaskRevoke.as_view()),
    # Events WebSocket API
    # url(r"api/task/events/task-sent/(.*)", events.TaskSent),
    # url(r"api/task/events/task-received/(.*)", events.TaskReceived),
    # url(r"api/task/events/task-started/(.*)", events.TaskStarted),
    # url(r"api/task/events/task-succeeded/(.*)", events.TaskSucceeded),
    # url(r"api/task/events/task-failed/(.*)", events.TaskFailed),
    # url(r"api/task/events/task-revoked/(.*)", events.TaskRevoked),
    # url(r"api/task/events/task-retried/(.*)", events.TaskRetried),
    # url(r"api/task/events/task-custom/(.*)", events.TaskCustom),
    # # WebSocket Updates
    # url(r"update-dashboard", DashboardUpdateHandler),
    # # Monitors
    url(r"^monitor/$", monitor.Monitor.as_view(), name='monitor'),
    url(r"^monitor/succeeded-tasks/$", monitor.SucceededTaskMonitor.as_view()),
    url(r"^monitor/failed-tasks/$", monitor.FailedTaskMonitor.as_view()),
    url(r"^monitor/completion-time/$", monitor.TimeToCompletionMonitor.as_view()),
    url(r"^monitor/broker/$", monitor.BrokerMonitor.as_view()),
    # Error
    url(r"^.*/$", NotFoundErrorHandler.as_view()),
], 'flower')


urlpatterns = [url("^", include(ns_urlpatterns))]


from __future__ import absolute_import

from django.http import Http404

from ..views import BaseHandler


class NotFoundErrorHandler(BaseHandler):
    def get(self, request):
        raise Http404

    def post(self, request):
        raise Http404



from mock import Mock, patch
from datetime import datetime, timedelta

from celery.result import AsyncResult
import celery.states as states

from tests.unit import AsyncHTTPTestCase


class ApplyTests(AsyncHTTPTestCase):
    def test_apply(self):
        from mock import patch, PropertyMock
        import json

        result = 'result'
        with patch('celery.result.AsyncResult.state', new_callable=PropertyMock) as mock_state:
            with patch('celery.result.AsyncResult.result', new_callable=PropertyMock) as mock_result:
                mock_state.return_value = states.SUCCESS
                mock_result.return_value = result

                ar = AsyncResult(123)
                ar.get = Mock(return_value=result)

                task = self._app.capp.tasks['foo'] = Mock()
                task.apply_async = Mock(return_value=ar)

                r = self.post('/api/task/apply/foo', body='')

        self.assertEqual(200, r.code)
        body = bytes.decode(r.body)
        self.assertEqual(result, json.loads(body)['result'])
        task.apply_async.assert_called_once_with(args=[], kwargs={})


class AsyncApplyTests(AsyncHTTPTestCase):
    def test_async_apply(self):
        task = self._app.capp.tasks['foo'] = Mock()
        task.apply_async = Mock(return_value=AsyncResult(123))
        r = self.post('/api/task/async-apply/foo', body={})

        self.assertEqual(200, r.code)
        task.apply_async.assert_called_once_with(args=[], kwargs={})

    def test_async_apply_eta(self):
        task = self._app.capp.tasks['foo'] = Mock()
        task.apply_async = Mock(return_value=AsyncResult(123))
        tomorrow = datetime.utcnow() + timedelta(days=1)
        r = self.post('/api/task/async-apply/foo',
                      body='{"eta": "%s"}' % tomorrow)

        self.assertEqual(200, r.code)
        task.apply_async.assert_called_once_with(
            args=[], kwargs={}, eta=tomorrow)

    def test_async_apply_countdown(self):
        task = self._app.capp.tasks['foo'] = Mock()
        task.apply_async = Mock(return_value=AsyncResult(123))
        r = self.post('/api/task/async-apply/foo',
                      body='{"countdown": "3"}')

        self.assertEqual(200, r.code)
        task.apply_async.assert_called_once_with(
            args=[], kwargs={}, countdown=3)

    def test_async_apply_expires(self):
        task = self._app.capp.tasks['foo'] = Mock()
        task.apply_async = Mock(return_value=AsyncResult(123))
        r = self.post('/api/task/async-apply/foo',
                      body='{"expires": "60"}')

        self.assertEqual(200, r.code)
        task.apply_async.assert_called_once_with(
            args=[], kwargs={}, expires=60)

    def test_async_apply_expires_datetime(self):
        task = self._app.capp.tasks['foo'] = Mock()
        task.apply_async = Mock(return_value=AsyncResult(123))
        tomorrow = datetime.utcnow() + timedelta(days=1)
        r = self.post('/api/task/async-apply/foo',
                      body='{"expires": "%s"}' % tomorrow)

        self.assertEqual(200, r.code)
        task.apply_async.assert_called_once_with(
            args=[], kwargs={}, expires=tomorrow)


class MockTasks:

    @staticmethod
    def get_task_by_id(events, task_id):
        from celery.events.state import Task
        return Task()


class TaskTests(AsyncHTTPTestCase):

    @patch('flower.api.tasks.tasks', new=MockTasks)
    def test_task_info(self):
        self.get('/api/task/info/123')



from __future__ import absolute_import

from collections import defaultdict

from celery import states
from flower.utils import login_required_admin
from django.utils.decorators import method_decorator

from ..views import BaseHandler
from ..utils.broker import Broker
from ..api.control import ControlHandler


class Monitor(BaseHandler):

    @method_decorator(login_required_admin)
    def get(self, request):
        return self.render("flower/monitor.html")


class SucceededTaskMonitor(BaseHandler):

    @method_decorator(login_required_admin)
    def get(self, request):
        timestamp = self.get_argument('lastquery', type=float)

        state = self.settings.state

        data = defaultdict(int)
        for task_key in state.tasks.keys():
            task = state.tasks[task_key]
            if timestamp < task.timestamp and task.state == states.SUCCESS:
                data[task.worker.hostname] += 1
        for worker_key in state.workers.keys():
            if worker_key not in data:
                data[worker_key] = 0

        return self.write(data)


class TimeToCompletionMonitor(BaseHandler):

    @method_decorator(login_required_admin)
    def get(self, request):
        timestamp = self.get_argument('lastquery', type=float)
        state = self.settings.state

        execute_time = 0
        queue_time = 0
        num_tasks = 0
        for task_key in state.tasks.keys():
            task = state.tasks[task_key]
            if timestamp < task.timestamp and task.state == states.SUCCESS:
                # eta can make "time in queue" look really scary.
                if task.eta is not None:
                    continue

                if task.started is None or task.received is None or\
                        task.succeeded is None:
                    continue

                queue_time += task.started - task.received
                execute_time += task.succeeded - task.started
                num_tasks += 1

        avg_queue_time = (queue_time / num_tasks) if num_tasks > 0 else 0
        avg_execution_time = (execute_time / num_tasks) if num_tasks > 0 else 0

        result = {
            "Time in a queue": avg_queue_time,
            "Execution time": avg_execution_time,
        }
        return self.write(result)


class FailedTaskMonitor(BaseHandler):

    @method_decorator(login_required_admin)
    def get(self, request):
        timestamp = self.get_argument('lastquery', type=float)
        state = self.settings.state

        data = defaultdict(int)
        for task_key in state.tasks.keys():
            task = state.tasks[task_key]
            if timestamp < task.timestamp and task.state == states.FAILURE:
                data[task.worker.hostname] += 1
        for worker_key in state.workers.keys():
            if worker_key not in data:
                data[worker_key] = 0

        return self.write(data)


class BrokerMonitor(BaseHandler):

    @method_decorator(login_required_admin)
    def get(self, request):
        app = self.settings.app
        try:
            broker = Broker(app.connection().as_uri(include_password=True),
                            http_api=self.settings.broker_api)
        except NotImplementedError:
            return self.write({})

        queue_names = ControlHandler.get_active_queue_names()
        queues = broker.queues(queue_names)

        data = defaultdict(int)
        for queue in queues:
            data[queue['name']] = queue.get('messages', 0)

        return self.write(data)



from __future__ import absolute_import

import logging
import sys

import kombu.exceptions
from flower.utils import login_required_admin
from django.utils.decorators import method_decorator

from ..api.control import ControlHandler
from ..utils.broker import Broker
from ..views import BaseHandler

logger = logging.getLogger(__name__)


class BrokerView(BaseHandler):

    @method_decorator(login_required_admin)
    def get(self, request):
        app = self.capp
        broker_options = app.conf.BROKER_TRANSPORT_OPTIONS
        http_api = None

        from celery.backends.amqp import AMQPBackend
        if isinstance(app.backend, AMQPBackend) and app.options.broker_api:
            http_api = app.options.broker_api

        try:
            broker = Broker(app.connection().as_uri(include_password=True),
                            http_api=http_api, broker_options=broker_options)
        except NotImplementedError:
            return self.write_error(404, message="'%s' broker is not supported" % app.transport)

        # noinspection PyBroadException
        try:
            queue_names = ControlHandler.get_active_queue_names()
            if not queue_names:
                queue_names = {app.conf.CELERY_DEFAULT_QUEUE} | \
                              set([q.name for q in app.conf.CELERY_QUEUES or [] if q.name])
            queues = list(broker.queues(sorted(queue_names)))

        except kombu.exceptions.OperationalError as e:
            return self.write_error(500, message="Unable to connect to broker",
                                    exc_info=sys.exc_info())
        except Exception:
            return self.write_error(500, message="Unable to get queues",
                                    exc_info=sys.exc_info())

        return self.render("flower/broker.html",
                           context={
                               'broker_url': app.connection().as_uri(),
                               'queues': queues
                           })
